import os
from collections import namedtuple

import pygame
from pygame.locals import *
import numpy as np

pygame.init()

# set up colors
colors_dict = {
    'black': (0, 0, 0),
    'white': (255, 255, 255),
    'red': (255, 0, 0),
    'green': (0, 255, 0),
    'blue': (0, 0, 255),
}


class AbstractStimMaker:
    """parent class for all StimMaker classes
    """

    RectTuple = namedtuple('RectTuple', ['display_surface',
                                         'grid_as_char',
                                         'target_indices',
                                         'distractor_indices'])

    def __init__(self,
                 target_color='red',
                 distractor_color='green',
                 window_size=(227, 227),
                 grid_size=(5, 5),
                 rects_width_height=(10, 30),
                 jitter=5):
        """__init__ function for Stim Makers

        Parameters
        ----------
        target_color : str
            {'red', 'green', 'white', 'blue'}. Default is 'red'.
        distractor_color : str
            {'red', 'green', 'white', 'blue'}. Default is 'green'.
        window_size : tuple
            of length two, representing (width, height) of window in pixels.
        grid_size : tuple
            of length two, representing the number of (rows, columns)
            in the grid on which tareget and distractors will be located.
        rects_width_height : tuple
            shape of pygame Rect objects that will be plotted on grid,
            (width, height) in pixels. Default is (10, 30).
        jitter : int
            number of pixels to jitter each 'item' in the 'set'
            that will be plotted on the grid. Default is 5.
            Adding jitter is helpful for creating stimuli with the
            same set size but slightly different placements, e.g. for
            augmenting data to train a learning algorithm and encourage
            invariant representations.
        """
        if not all([type(grid_size_el) == int for grid_size_el in grid_size]):
            raise ValueError('values for grid size must be positive integers')

        if not all([grid_size_el > 0 for grid_size_el in grid_size]):
            raise ValueError('values for grid size must be positive integers')

        self.target_color = target_color
        self.distractor_color = distractor_color
        self.grid_size = grid_size
        self.window_size = window_size
        self.rects_width_height = rects_width_height
        self.jitter = jitter

    def _return_rect_for_stim(self, display_surface, rect_to_draw, is_target):
        """this function must return a pygame.rect object
        which the StimMaker.make_stim method will `blit` on
        the stimulus."""
        raise NotImplementedError

    def make_stim(self,
                  set_size=8,
                  num_target=1,
                  ):
        """make visual search stimuli

        Parameters
        ----------
        set_size : int
            Set size, equal to number of targets + distractors.
            Default is 8.
        num_target : int
            Number of targets. Default is 1.

        Returns
        -------
        display_surface : pygame.Surface
            with visual search stimuli plotted on it
        """
        if type(set_size) != int:
            raise TypeError('set size must be an integer')

        if not set_size > 0:
            raise ValueError('set size must be greater than zero')

        if type(num_target) != int:
            raise TypeError('number of targets must be an integer')

        if not num_target >= 0:
            raise ValueError('number of targets must be greater than or equal to zero')

        if num_target > set_size:
            raise ValueError('number of targets cannot be greater than set size')

        total_grid_elements = self.grid_size[0] * self.grid_size[1]
        if set_size > total_grid_elements:
            raise ValueError('set size {} cannot be greater than number of '
                             'elements in grid, {}'
                             .format(set_size, total_grid_elements))

        if type(self.jitter) != int:
            raise TypeError('value for jitter must be an integer')

        # make grid, randomly select which cells in grid to use
        grid_x = np.arange(1, self.grid_size[0] + 1)
        grid_y = np.arange(1, self.grid_size[1] + 1)
        xx, yy = np.meshgrid(grid_x, grid_y)
        xx = xx.ravel()
        yy = yy.ravel()
        num_cells = self.grid_size[0] * self.grid_size[1]
        cells_to_use = sorted(np.random.choice(np.arange(num_cells),
                                               size=set_size,
                                               replace=False))

        xx_to_use = xx[cells_to_use]
        yy_to_use = yy[cells_to_use]

        # find centers of cells we're going to use
        cell_width = round(self.window_size[0] / self.grid_size[0])
        cell_x_center = round((self.window_size[0] / self.grid_size[0]) / 2)
        xx_to_use_ctr = (xx_to_use * cell_width) - cell_x_center
        cell_height = round(self.window_size[1] / self.grid_size[1])
        cell_y_center = round((self.window_size[1] / self.grid_size[1]) / 2)
        yy_to_use_ctr = (yy_to_use * cell_height) - cell_y_center

        # add jitter to those points
        jitter_high = self.jitter // 2
        jitter_low = -jitter_high
        if self.jitter % 2 == 0:  # if even
            # have to account for zero at center of jitter range
            # (otherwise range would be jitter + 1)
            # (Not a problem when doing floor division on odd #s)
            coin_flip = np.random.choice([0, 1])
            if coin_flip == 0:
                jitter_low += 1
            elif coin_flip == 1:
                jitter_high -= 1
        jitter_range = np.arange(jitter_low, jitter_high + 1)

        x_jitter = np.random.choice(jitter_range, size=xx_to_use.size)
        xx_to_use_ctr += x_jitter
        y_jitter = np.random.choice(jitter_range, size=yy_to_use.size)
        yy_to_use_ctr += y_jitter

        # set up window
        display_surface = pygame.display.set_mode(self.window_size, 0, 32)

        # draw on surface object
        display_surface.fill(colors_dict['black'])
        target_inds = np.random.choice(np.arange(set_size),
                                       size=num_target).tolist()
        rects = []
        target_indices = []
        distractor_indices = []
        grid_as_char = [''] * num_cells
        for item in range(set_size):
            rect_tuple = (0, 0) + self.rects_width_height
            rect_to_draw = Rect(rect_tuple)
            center = (int(xx_to_use_ctr[item]), int(yy_to_use_ctr[item]))
            rect_to_draw.center = center
            if item in target_inds:
                is_target = True
                target_indices.append(center)
                grid_as_char[cells_to_use[item]] = 't'
            else:
                is_target = False
                distractor_indices.append(center)
                grid_as_char[cells_to_use[item]] = 'd'
            curr_rect = self._return_rect_for_stim(display_surface=display_surface,
                                                   rect_to_draw=rect_to_draw,
                                                   is_target=is_target)
            rects.append(curr_rect)
        grid_as_char = np.asarray(grid_as_char).reshape(self.grid_size[0], self.grid_size[1]).tolist()
        pygame.display.update()
        return self.RectTuple(display_surface=display_surface,
                              grid_as_char=grid_as_char,
                              target_indices=target_indices,
                              distractor_indices=distractor_indices)


class RectangleStimMaker(AbstractStimMaker):
    """Make visual search stimuli with vertical rectangles
    where target is different color form distractor.
    Considered a stimulus that allows for 'efficient' search."""
    def _return_rect_for_stim(self, display_surface, rect_to_draw, is_target):
        if is_target:
            color = self.target_color
        else:
            color = self.distractor_color
        rect = pygame.draw.rect(display_surface, colors_dict[color], rect_to_draw)
        return rect


class NumberStimMaker(AbstractStimMaker):
    """Make visual search stimuli with digital 2s and 5s.
    Considered a stimulus that results in 'inefficient' search."""
    png_path = os.path.join(os.path.dirname(__file__), 'png')

    numbers_dict = {
        (2, 'red'): 'two_red.png',
        (2, 'green'): 'two_green.png',
        (2, 'white'): 'two_white.png',
        (5, 'red'): 'five_red.png',
        (5, 'green'): 'five_green.png',
        (5, 'white'): 'five_white.png',
    }

    def __init__(self,
                 target_color='white',
                 distractor_color='white',
                 grid_size=(5, 5),
                 window_size=(227, 227),
                 rects_width_height=(30, 30),
                 jitter=5,
                 target_number=2,
                 distractor_number=5):
        """
        Other Parameters
        ----------------
        target_number : int
            Number that is used as target.
            one of {2, 5}. Default is 2.
        distractor_number : int
            Number that is used as a distractor.
            one of {2, 5}. Default is 5.
        """
        super().__init__(target_color=target_color,
                         distractor_color=distractor_color,
                         window_size=window_size,
                         grid_size=grid_size,
                         rects_width_height=rects_width_height,
                         jitter=jitter)
        self.target_number = target_number
        self.distractor_number = distractor_number
        self.target_png = os.path.join(self.png_path,
                                       self.numbers_dict[(self.target_number,
                                                          self.target_color)])
        self.target = pygame.image.load(self.target_png)
        self.distractor_png = os.path.join(self.png_path,
                                           self.numbers_dict[(self.distractor_number,
                                                              self.distractor_color)])
        self.distractor = pygame.image.load(self.distractor_png)

    def _return_rect_for_stim(self, display_surface, rect_to_draw, is_target):
        """Returns target or distractor"""
        if is_target:
            rect = display_surface.blit(self.target, rect_to_draw)
        else:
            rect = display_surface.blit(self.distractor, rect_to_draw)
        return rect



import argparse
import os

from .config import parse
from .make import make


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('configfile',
                        type=str,
                        help=('filename of config.ini file, e.g.:\n'
                              '$ searchstims ./basic_config.ini\n'
                              'For an example config.ini file, see: '
                              'https://github.com/NickleDave/searchstims'))
    args = parser.parse_args()
    config_file = args.configfile
    if not os.path.isfile(config_file):
        raise FileNotFoundError("Config file {} not found".format(config_file))
    config_tuple = parse(config_file)
    make(config_tuple)


if __name__ == '__main__':
    main()



import os
import json
import pygame

from . import stim_makers


def make(config_tuple):
    """actual function that makes all the stimuli

    Parameters
    ----------
    config_tuple : ConfigTuple
        returned by parse_config after parsing a config.ini file

    Returns
    -------
    None

    saves all the stimuli to config_tuple.output_dir, and saves information about
    stimuli in a .json output file. This .json file is a serialized Python
    dictionary of dictionaries with the following key, field pairs:
    {set size: {
        'present': [
            {'filename': str,
             'grid_as_char': list,
            ]

    Keys at the top level are set size, the total number of targets and distractors, e.g.,
    {1, 2, ..., 8}. Each set size key has as its value another dictionary,
    whose keys are 'present' and 'absent', referring to the visual search target.
    Each 'present' and 'absent' key has as its value a list of Python dictionaries;
    each dictionary in the list has info about the actual visual search stimulus image
    that it corresponds to:
        filename: str
            actual visual search stimulus filename
        grid_as_char: list
            of list of str. Representation of stimulus as a grid of
            cells
        target_indices: list
            of two-element lists, the x and y co-ordinates for the
            center of the targets (or indices if you load the image into
            an array).
        distractor_indices: list
            of two-element lists, the x and y co-ordinates for the
            center of the distractors (or indices if you load the image into
            an array).

    Here is an excerpt from such a file:
        {'1': {'absent': [{'distractor_indices': [[203, 65]],
            'filename': '/home/user/output/1/absent/redvert_v_greenvert_set_size_1_target_absent_0.png',
            'grid_as_char':
                [['', '', '', '', ''],
                 ['', '', '', '', 'd'],
                 ['', '', '', '', ''],
                 ['', '', '', '', ''],
                 ['', '', '', '', '']],
            'target_indices': []},
           {'distractor_indices': [[111, 21]],
            'filename': '/home/user/output/1/absent/redvert_v_greenvert_set_size_1_target_absent_1.png',
            ...
         '2': {'absent': [{'distractor_indices': [[68, 22], [65, 204]],
            'filename': '/home/user/output/2/absent/redvert_v_greenvert_set_size_2_target_absent_0.png',
            'grid_as_char': [['', 'd', '', '', ''],
            ...
    """
    if not os.path.isdir(config_tuple.general.output_dir):
        os.makedirs(config_tuple.general.output_dir)
    # put filenames and other info in a dict that we serialize as json
    # so we don't have to do a bunch of string matching to find filenames later,
    # instead we just load back into Python as a dict 
    # and can just get all the filenames for a given set size with target present or absent
    # by using appropriate keys
    # e.g. fnames_set_size_8_target_present = [stim_info['filename'] for stim_info in out_dict[8]['present']]
    out_dict = {}

    if hasattr(config_tuple, 'rectangle'):
        init_config = config_tuple.rectangle
        stim_maker = stim_makers.RectangleStimMaker(target_color=init_config.target_color,
                                                    distractor_color=init_config.distractor_color,
                                                    window_size=init_config.image_size,
                                                    grid_size=init_config.grid_size,
                                                    rects_width_height=init_config.rects_width_height,
                                                    jitter=init_config.jitter
                                                    )
    elif hasattr(config_tuple, 'number'):
        init_config = config_tuple.number
        stim_maker = stim_makers.NumberStimMaker(target_color=init_config.target_color,
                                                 distractor_color=init_config.distractor_color,
                                                 window_size=init_config.image_size,
                                                 grid_size=init_config.grid_size,
                                                 rects_width_height=init_config.rects_width_height,
                                                 jitter=init_config.jitter,
                                                 target_number=init_config.target_number,
                                                 distractor_number=init_config.distractor_number
                                                 )

    general_config = config_tuple.general
    for set_size in general_config.set_sizes:
        # add dict for this set size that will have list of "target present / absent" filenames
        out_dict[set_size] = {}

        if not os.path.isdir(
            os.path.join(general_config.output_dir, str(set_size))
        ):
            os.makedirs(
                os.path.join(general_config.output_dir, str(set_size))
            )
        for target in ('present', 'absent'):
            # add the actual filename list for 'present' or 'absent'
            out_dict[set_size][target] = []
            if target == 'present':
                inds_of_stim_to_make = range(general_config.num_target_present // len(general_config.set_sizes))
                num_target = 1
            elif target == 'absent':
                inds_of_stim_to_make = range(general_config.num_target_absent // len(general_config.set_sizes))
                num_target = 0

            if not os.path.isdir(
                    os.path.join(general_config.output_dir, str(set_size), target)
            ):
                os.makedirs(os.path.join(general_config.output_dir, str(set_size), target))

            for i in inds_of_stim_to_make:
                rect_tuple = stim_maker.make_stim(set_size=set_size,
                                                  num_target=num_target)
                if hasattr(config_tuple, 'rectangle'):
                    filename = ('redvert_v_greenvert_set_size_{}_'
                                'target_{}_{}.png'.format(set_size, target, i))
                elif hasattr(config_tuple, 'number'):
                    filename = ('two_v_five_set_size_{}_'
                                'target_{}_{}.png'.format(set_size, target, i))
                filename = os.path.join(general_config.output_dir,
                                        str(set_size),
                                        target,
                                        filename)
                pygame.image.save(rect_tuple.display_surface, filename)
                stim_info = {
                    'filename': filename,
                    'grid_as_char': rect_tuple.grid_as_char,
                    'target_indices': rect_tuple.target_indices,
                    'distractor_indices': rect_tuple.distractor_indices,
                }
                out_dict[set_size][target].append(stim_info)

    out_json = json.dumps(out_dict, indent=4)
    json_filename = os.path.expanduser(general_config.json_filename)
    with open(json_filename, 'w') as json_fp:
        print(out_json, file=json_fp)



import os
import configparser
from collections import namedtuple
from distutils.util import strtobool
import ast

this_file_dir = os.path.dirname(__file__)

CONFIG_TYPES = configparser.ConfigParser()
CONFIG_TYPES.read(os.path.join(this_file_dir, 'types.ini'))
DEFAULT_CONFIG = configparser.ConfigParser()
DEFAULT_CONFIG.read(os.path.join(this_file_dir, 'default.ini'))


def parse(config_file=None, config=None):
    """read config.ini file with config parser,
    returns namedtuple ConfigTuple with
    sections and options as attributes.

    Parameters
    ----------
    config_file : str
        Path to a config.ini file. If None, the
        default configuration is returned.
        Default is None.
    config: configparser.ConfigParser
        instance of ConfigParser that already has sections and options declared.
        This parameter is mainly used in testing, to write tests where the options
        are explicit in that file.

    Returns
    -------
    config_tuple : namedtuple
        where fields are sections of the config, and
        values for those fields are also namedtuples,
        with fields being options and values being the
        values for those options from the config.ini file.
    """
    if config_file and config:
        raise TypeError('cannot call parse function with config_file and config, '
                        'unclear which to use.')

    if config_file:
        if not os.path.isfile(config_file):
            raise FileNotFoundError(f'config_file {config_file} not found')

        config = configparser.ConfigParser()
        config.read(config_file)

    config_sections = config.sections()
    if len(set(config_sections) - {'general'}) > 1:
        raise ValueError('config has more than one section with options for visual search stimuli, '
                         'unclear which to use for generating the stimuli.')

    for section in config.sections():
        for option in DEFAULT_CONFIG.options(section):
            if not config.has_option(section, option):
                config[section][option] = DEFAULT_CONFIG[section][option]

    if config.has_option(section='general',
                         option='output_dir'):
        output_dir = config['general']['output_dir']
        config['general']['output_dir'] = os.path.expanduser(output_dir)
    else:
        config['general']['output_dir'] = os.path.join('.', 'output')

    if config.has_option(section='general',
                         option='json_filename'):
        if os.path.split(config['general']['json_filename'])[0] == '':
            config['general']['json_filename'] = os.path.join(output_dir,
                                                             config['general']['json_filename'])
        else:
            config['general']['json_filename'] = config['general']['json_filename']
    else:
        # default filename if option not used
        config['general']['json_filename'] = os.path.join(output_dir,
                                                          'filenames_by_set_size_'
                                                          'and_target.json')

    # fancy way of turning ConfigParser instance into a namedtuple
    sections = [key for key in list(config.keys()) if key != 'DEFAULT']
    ConfigTuple = namedtuple('ConfigTuple', sections)
    config_dict = {}
    for section in sections:
        section_keys = list(config[section].keys())
        section_values = list(config[section].values())
        SubTup = namedtuple(section, section_keys)
        subtup_dict = {}
        for key, val in zip(section_keys, section_values):
            val_type = CONFIG_TYPES[section][key]
            if val_type == 'int':
                typed_val = int(val)
            elif val_type == 'float':
                typed_val = float(val)
            elif val_type == 'bool':
                typed_val = bool(strtobool(val))
            elif val_type == 'list' or val_type == 'tuple':
                typed_val = ast.literal_eval(val)
            elif val_type == 'str':
                typed_val = val
            subtup_dict[key] = typed_val
        config_dict[section] = SubTup(**subtup_dict)
    config_tuple = ConfigTuple(**config_dict)
    return config_tuple
